@@ -31,7 +31,7 @@ import datetime
 import time
 import collections
 
-VERSION = (0, 3, 4)
+VERSION = (0, 3, 5)
 __version__ = '%s.%s.%s' % VERSION
 apilevel = '2.0'
 threadsafety = 1
@@ -124,6 +124,7 @@ PG_TYPE_CASH = 790
 PG_TYPE_MACADDR = 829
 PG_TYPE_INET = 869
 PG_TYPE_CIDR = 650
+PG_TYPE_NAMEARRAY = 1003
 PG_TYPE_INT2ARRAY = 1005
 PG_TYPE_INT4ARRAY = 1007
 PG_TYPE_TEXTARRAY = 1009
@@ -174,15 +175,28 @@ PG_TYPE_FDW_HANDLER = 3115
 PG_TYPE_ANYRANGE = 3831
 
 def _decode_column(data, oid, encoding):
-    class UTC(datetime.tzinfo):
+    class TZ(datetime.tzinfo):
+        def __init__(self, offset):
+            self.offset = offset
+            hours = int(self.offset[:3])
+            if len(offset) > 3:
+                minutes = int(self.offset[4:6])
+            else:
+                minutes = 0
+            if len(offset) > 6:
+                seconds = int(self.offset[7:])
+            else:
+                seconds = 0
+            self.delta = datetime.timedelta(hours=hours, minutes=minutes, seconds=seconds)
+
         def utcoffset(self, dt):
-            return datetime.timedelta(0)
+            return self.delta
 
         def dst(self, dt):
-            return datetime.timedelta(0)
+            return self.delta
 
         def tzname(self, dt):
-            return "UTC"
+            return u"UTC" + self.offset
 
     if data is None:
         return data
@@ -219,22 +233,18 @@ def _decode_column(data, oid, encoding):
             dt = datetime.datetime.strptime(s, '%H:%M:%S')
         else:
             dt = datetime.datetime.strptime(s, '%H:%M:%S.%f')
-        dt.replace(tzinfo=UTC())
-        dt += datetime.timedelta(hours=offset)
+        dt.replace(tzinfo=TZ(data[n:]))
         return datetime.time(dt.hour, dt.minute, dt.second, dt.microsecond, tzinfo=dt.tzinfo)
     elif oid in (PG_TYPE_TIMESTAMPTZ, ):
         n = data.rfind('+')
         if n == -1:
             n = data.rfind('-')
         s = data[:n]
-        offset = int(data[n:])
         if len(s) == 19:
             dt = datetime.datetime.strptime(s, '%Y-%m-%d %H:%M:%S')
         else:
             dt = datetime.datetime.strptime(s, '%Y-%m-%d %H:%M:%S.%f')
-        dt.replace(tzinfo=UTC())
-        dt += datetime.timedelta(hours=offset)
-        return dt
+        return dt.replace(tzinfo=TZ(data[n:]))
     elif oid in (PG_TYPE_INTERVAL, ):
         dt = data.split('days')
         if len(dt) < 2:
@@ -258,13 +268,17 @@ def _decode_column(data, oid, encoding):
         return b''.join([chr(c) for c in ia]) if PY2 else bytes(ia)
     elif oid in (PG_TYPE_TEXT, PG_TYPE_BPCHAR, PG_TYPE_VARCHAR, PG_TYPE_NAME, PG_TYPE_JSON):
         return data
-    elif oid in (PG_TYPE_UNKNOWN, PG_TYPE_PGNODETREE, PG_TYPE_UUID, PG_TYPE_TSVECTOR):
+    elif oid in (PG_TYPE_UNKNOWN, PG_TYPE_PGNODETREE, PG_TYPE_UUID, PG_TYPE_TSVECTOR, PG_TYPE_INET):
         if DEBUG: DEBUG_OUTPUT('NO DECODE type:%d' % (oid, ))
         return data
     elif oid in (PG_TYPE_INT2ARRAY, PG_TYPE_INT4ARRAY):
         return [int(i) for i in data[1:-1].split(',')]
+    elif oid in (PG_TYPE_NAMEARRAY, PG_TYPE_TEXTARRAY):
+        return [s for s in data[1:-1].split(',')]
     elif oid in (PG_TYPE_FLOAT4ARRAY, ):
         return [float(f) for f in data[1:-1].split(',')]
+    elif oid in (PG_TYPE_INT2VECTOR, ):
+        return [int(i) for i in data.split(' ')]
     else:
         if DEBUG:
             raise ValueError('Unknown oid=' + str(oid) + ":" + data)
@@ -284,8 +298,16 @@ def _bint_to_bytes(val):    # Convert int value to big endian 4 bytes.
     else:
         return bytes([(val >> 24) & 0xff, (val >> 16) & 0xff, (val >> 8) & 0xff, val & 0xff])
 
+_escape_hook = {}
+
+def add_escape_hook(t, func):
+    _escape_hook[t] = func
+
 def escape_parameter(v):
     t = type(v)
+    func = _escape_hook.get(t)
+    if func:
+        return func(v)
     if v is None:
         return 'NULL'
     elif (PY2 and t == unicode) or (not PY2 and t == str):  # string
@@ -300,6 +322,10 @@ def escape_parameter(v):
     elif t == time.struct_time:
         return u'%04d-%02d-%02d %02d:%02d:%02d' % (
             v.tm_year, v.tm_mon, v.tm_mday, v.tm_hour, v.tm_min, v.tm_sec)
+    elif t == datetime.datetime:
+        return "'" + str(v) + "'"
+    elif t == datetime.date:
+        return "'" + str(v) + "'"
     elif t == datetime.timedelta:
         if v.seconds:
             return u"interval '%d second'" % (v.days * 86400 + v.seconds, )
@@ -314,7 +340,6 @@ def escape_parameter(v):
     else:
         return "'" + str(v) + "'"
 
-
 Date = datetime.date
 Time = datetime.time
 TimeDelta = datetime.timedelta
@@ -388,6 +413,7 @@ class Cursor(object):
         self._rows = collections.deque()
         self._rowcount = 0
         self.arraysize = 1
+        self.query = None
 
     def callproc(self, procname, args=()):
         raise NotSupportedError()
@@ -403,6 +429,8 @@ class Cursor(object):
 
     def execute(self, query, args=()):
         if DEBUG: DEBUG_OUTPUT('Cursor::execute()\t%s\t%s' % (query, args))
+        if not self.connection or not self.connection.is_connect():
+            raise ProgrammingError(u"08003:Lost connection")
         self.description = []
         self._rows.clear()
         self.query = query
@@ -411,6 +439,7 @@ class Cursor(object):
             escaped_args = tuple(escape_parameter(arg) for arg in args)
             query = query.replace('%', '%%').replace('%%s', '%s')
             query = query % escaped_args
+            query = query.replace('%%', '%')
         self.connection.execute(query, self)
 
     def executemany(self, query, seq_of_params):
@@ -450,6 +479,10 @@ class Cursor(object):
     def rowcount(self):
         return self._rowcount
 
+    @property
+    def closed(self):
+        return self.connection is None or not self.connection.is_connect()
+
     def __iter__(self):
         return self
 
@@ -673,10 +706,7 @@ class Connection(object):
     def cursor(self):
         return Cursor(self)
 
-    def execute(self, query, obj=None):
-        if not self.in_transaction:
-            self.begin()
-        if DEBUG: DEBUG_OUTPUT('Connection::execute()\t%s' % (query, ))
+    def _execute(self, query, obj=None):
         self._send_message(
             PG_F_QUERY,
             query.encode(self.encoding) + b'\x00',
@@ -685,6 +715,12 @@ class Connection(object):
         if self.autocommit:
             self.commit()
 
+    def execute(self, query, obj=None):
+        if DEBUG: DEBUG_OUTPUT('Connection::execute()\t%s' % (query, ))
+        if not self.in_transaction:
+            self.begin()
+        self._execute(query, obj)
+
     def set_autocommit(self, autocommit):
         self.autocommit = autocommit
 
